/*
** Module   : AppServerHelper.java
** Abstract : A helper class to ease appserver connection and remote calls. 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130529 Created initial version.
** 002 CA  20130813 Track the mode of the connection (session-free or session-managed). Implement
**                  async calls.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.Array;
import java.net.*;
import java.sql.*;

import com.goldencode.p2j.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.persist.*;

/**
 * Provides a collection of APIs for connecting to the appserver and performing remote calls.
 * <p>
 * When the {@link #connect} or {@link #connectLegacyMode} is called, an instance of this class is
 * returned which contains the connection ID and can be used to execute remote calls, via the
 * determined {@link #appServer} network server.
 * <p>
 * When establishing connections from Java code, use the {@link #connect}, {@link #invoke} and
 * {@link #invokePersistent} APIs to establish a connection and send requests to the remote side.
 * The {@link #getConnectError()}, {@link #getConnectResult()} and {@link #getConnectionID()} APIs
 * provide information about the state of the last CONNECT request; note that a connection ID will
 * be sent back only if the connection was successful.
 */
public class AppServerHelper
{
   /**
    * A local appserver instance, used when the P2J server running the appserver is the same as
    * the P2J server from where the connection request originates.
    */
   private static AppServerEntry localAppServer = null;

   /** The id of this appserver connection. */
   private final String connectionID;

   /** The name of the connected appserver. */
   private final String appServerName;
   
   /**
    * Flag indicating if this connection is in session free mode (if true) or in session-managed
    * mode (if false).
    */
   private final boolean sessionFree;
   
   /**
    * A network server to the P2J server running this appserver. Each call via this server
    * (except the connect-related calls) will have the connection ID as the first parameter.
    */
   private final AppServerEntry appServer;

   /** The result returned by this CONNECT call. */
   private character connectResult = null;
   
   /** The error encountered during the CONNECT call. */
   private NumberedException connectError = null;
   
   /**
    * The P2J session used to communicate with the remote side. May be <code>null</code> if the
    * appserver is in the same JVM as the requester. */
   private Session session = null;
   
   /** Flag indicating the appserver is running in the same JVM as the requester. */
   private boolean local = false;
   
   /**
    * Initialize this helper when the connection was not established.
    */
   private AppServerHelper()
   {
      this(null, null, null, null, false);
   }
   
   /**
    * Instantiate a new helper to communicate with the specified appServer instance.
    * 
    * @param    appServerName
    *           The name of the connected appserver. 
    * @param    connectionID
    *           The ID for this appServer connection.
    * @param    session
    *           The P2J session used to communicate with the remote side. May be <code>null</code>
    *           if the appServer is on the same P2J Server as the requester.
    * @param    appServer
    *           The {@link AppServerEntry} network server instance.
    */
   private AppServerHelper(String         appServerName,
                           String         connectionID,
                           Session        session,
                           AppServerEntry appServer,
                           boolean        sessionFree)
   {
      this.appServerName = appServerName;
      this.connectionID = connectionID;
      this.session = session;
      this.appServer = appServer;
      this.sessionFree = sessionFree;
   }

   /**
    * Authenticate with the given P2J server and establish an appserver connection.  If the
    * address is the address of this running P2J server, then a local proxy is resolved and no
    * P2J connection is established.
    * 
    * @param    config
    *           The configuration for the P2J server connection.
    * @param    account
    *           The P2J account to authenticate on the remote side.
    * @param    appServerName
    *           The name of the appserver to which the connection will be established.
    * @param    sessionFree
    *           <code>true</code> to indicate a session-free operating mode, <code>false</code>
    *           to indicate a session-managed operating mode.
    *
    * @return   The helper with the connection result (which was established or not).
    */
   public static AppServerHelper connect(BootstrapConfig  config,
                                         String           account,
                                         String           appServerName,
                                         boolean          sessionFree)
   {
      return connect(config, account, null, null, appServerName, sessionFree, null, null, null);
   }

   /**
    * Authenticate with the given P2J server and establish an appserver connection.  If the
    * address is the address of this running P2J server, then a local proxy is resolved and no
    * P2J connection is established.
    * 
    * @param    config
    *           The configuration for the P2J server connection.
    * @param    account
    *           The P2J account to authenticate on the remote side.
    * @param    appServerName
    *           The name of the appserver to which the connection will be established.
    * @param    sessionFree
    *           <code>true</code> to indicate a session-free operating mode, <code>false</code>
    *           to indicate a session-managed operating mode.
    * @param    user
    *           The username passed to the appserver's connect procedure.
    * @param    pwd
    *           The password passed to the appserver's connect procedure.
    * @param    serverInfo
    *           The info passed to the appserver's connect procedure.
    *
    * @return   The helper with the connection result (which was established or not).
    */
   public static AppServerHelper connect(BootstrapConfig   config,
                                         String            account,
                                         String            appServerName,
                                         boolean           sessionFree,
                                         String            user,
                                         String            pwd,
                                         String            serverInfo)
   throws NumberedException
   {
      return connect(config, account, null, null,
                     appServerName, sessionFree, user, pwd, serverInfo);
   }

   /**
    * Authenticate with the given P2J server and establish an appserver connection.  If the
    * address is the address of this running P2J server, then a local proxy is resolved and no
    * P2J connection is established.
    * 
    * @param    config
    *           The configuration for the P2J server connection.
    * @param    account
    *           The P2J account to authenticate on the remote side.
    * @param    listener
    *           The session listener, used when a virtual or direct session is established.
    * @param    handler
    *           The interrupt handler, used when a direct session is established.
    * @param    appServerName
    *           The name of the appserver to which the connection will be established.
    * @param    sessionFree
    *           <code>true</code> to indicate a session-free operating mode, <code>false</code>
    *           to indicate a session-managed operating mode.
    *
    * @return   The helper with the connection result (which was established or not).
    */
   public static AppServerHelper connect(BootstrapConfig  config,
                                         String           account,
                                         SessionListener  listener,
                                         InterruptHandler handler,
                                         String           appServerName,
                                         boolean          sessionFree)
   {
      return connect(config, account, listener, handler, appServerName, sessionFree,
                     null, null, null);
   }
   
   /**
    * Authenticate with the given P2J server and establish an appserver connection.  If the
    * address is the address of this running P2J server, then a local proxy is resolved and no
    * P2J connection is established.
    * 
    * @param    config
    *           The configuration for the P2J server connection.
    * @param    account
    *           The P2J account to authenticate on the remote side.
    * @param    listener
    *           The session listener, used when a virtual or direct session is established.
    * @param    handler
    *           The interrupt handler, used when a direct session is established.
    * @param    appServerName
    *           The name of the appserver to which the connection will be established.
    * @param    sessionFree
    *           <code>true</code> to indicate a session-free operating mode, <code>false</code>
    *           to indicate a session-managed operating mode.
    * @param    user
    *           The username passed to the appserver's connect procedure.
    * @param    pwd
    *           The password passed to the appserver's connect procedure.
    * @param    serverInfo
    *           The info passed to the appserver's connect procedure.
    *
    * @return   The helper with the connection result (which was established or not).
    */
   public static AppServerHelper connect(BootstrapConfig  config,
                                         String           account,
                                         SessionListener  listener,
                                         InterruptHandler handler,
                                         String           appServerName,
                                         boolean          sessionFree,
                                         String           user,
                                         String           pwd,
                                         String           serverInfo)
   {
      int port = -1;
      boolean isSecure = config.getBoolean("net", "connection", "secure", false);
      if (isSecure)
      {
         port = config.getInt("net", "server", "secure_port", -1);
      }
      else
      {
         port = config.getInt("net", "server", "port", -1);
         if (port == -1)
         {
            port = config.getInt("net", "server", "insecure_port", -1);
         }
      }
      String host = config.getString("net", "server", "host", null);
      InetSocketAddress remoteAddr = new InetSocketAddress(host, port);

      // when connecting, ignore all SSL-related options from the 4GL code, as they are
      // irrelevant
      Session session = null;
      SessionManager sessMgr = SessionManager.get();
      boolean local = false;
      
      if (sessMgr.isLeaf())
      {
         // this is a call originating from outside P2J code

         try
         {
            // establish a direct connection
            session = sessMgr.connectDirect(config, handler, listener);
         }
         catch (Exception e)
         {
            // the connection could not be established
            AppServerHelper helper = new AppServerHelper();
            helper.setConnectError(
                new NumberedException("Connection refused by Application Server", 5467, e));
            
            return helper;
         }
      }
      else
      {
         local = sessMgr.isLocalNode(remoteAddr);
         if (!local)
         {
            try
            {
               // establish a virtual connection
               session = sessMgr.connectVirtual(remoteAddr, null, listener, account);
            }
            catch (Exception e)
            {
               // the connection could not be established
               AppServerHelper helper = new AppServerHelper();
               helper.setConnectError(
                   new NumberedException("Connection refused by Application Server", 5467, e));
               
               return helper;
            }
         }
      }

      AppServerEntry appServer = null;
      if (session != null)
      {
         appServer = (AppServerEntry) RemoteObject.obtainNetworkInstance(AppServerEntry.class,
                                                                         session);
      }
      else
      {
         appServer = getLocalAppServer();
      }

      // the P2J account needs to be sent, as in local-proxy mode, a new context/authentication
      // is not performed, and all appserver requests would be sent from the current account.
      AppServerInvocationResult result = appServer.connect(appServerName,
                                                           sessionFree,
                                                           user,
                                                           pwd,
                                                           serverInfo);
      String connectionID = result.getConnectionID();
      AppServerHelper helper = new AppServerHelper(appServerName,
                                                   connectionID,
                                                   session,
                                                   appServer,
                                                   sessionFree);
      helper.local = local;

      if (result.getResult() != null)
      {
         helper.setConnectResult(new character((BaseDataType) result.getResult()));
      }
      
      if (result.getError() != null || connectionID == null)
      {
         // the connection could not be established
         helper.setConnectError(new NumberedException("Connection refused by Application Server",
                                                      5467, result.getError()));
      }
      
      return helper;
   }

   /**
    * This code is intended for legacy compatibility, and is used to connect to an appserver, when
    * a CONNECT statement is encountered.
    * <p>
    * This will use the known appservers configured in the directory to identify the P2J server
    * and account, on which the appserver targeted by the CONNECT statement is running.
    * 
    * @param    appService
    *           The value of the "-AppService" parameter.
    * @param    host
    *           The value of the "-H" parameter.
    * @param    serviceName
    *           The service name specified by the "-S" parameter.
    * @param    port
    *           The port specified by the "-S" parameter.
    * @param    sessionFree
    *           <code>true</code> to indicate a session-free operating mode, <code>false</code>
    *           to indicate a session-managed operating mode.
    * @param    user
    *           The username passed to the appserver's connect procedure.
    * @param    pwd
    *           The password passed to the appserver's connect procedure.
    * @param    info
    *           The info passed to the appserver's connect procedure.
    *
    * @return   The helper with the connection result (which was established or not).
    */
   static AppServerHelper connectLegacyMode(String    appService,
                                            String    host,
                                            String    serviceName,
                                            int       port,
                                            boolean   directConnect,
                                            boolean   sessionFree,
                                            character user,
                                            character pwd,
                                            character info)
   throws NumberedException
   {
      InetSocketAddress address = AppServerManager.lookupP2JServer(appService,
                                                                   host,
                                                                   serviceName,
                                                                   port,
                                                                   directConnect);
      if (address == null || address.getAddress() == null)
      {
         final String msg = "Connect failure for host %s port %s transport TCP";
         String sport = (serviceName == null ? Integer.toString(port) : serviceName);

         throw new NumberedException(String.format(msg, host, sport), 9407);
      }
      
      String p2jAccount = AppServerManager.lookupP2JAccount(appService,
                                                            host,
                                                            serviceName,
                                                            port,
                                                            directConnect);
      String appServerName = AppServerManager.lookupTargetAppserver(appService,
                                                                    host,
                                                                    serviceName,
                                                                    port,
                                                                    directConnect);
      
      String sUser = (user == null || user.isUnknown() ? null : user.toStringMessage());
      String sPwd  = (pwd  == null || pwd.isUnknown()  ? null : pwd.toStringMessage());
      String sInfo = (pwd  == null || info.isUnknown() ? null : info.toStringMessage());

      BootstrapConfig config = null;
      try
      {
         config = new BootstrapConfig();
         config.setConfigItem("net", "connection", "secure", "true");
         config.setConfigItem("net", "server", "host", address.getHostName());
         config.setConfigItem("net", "server", "secure_port", Integer.toString(address.getPort()));
      }
      catch (ConfigurationException e)
      {
         // this should never happen, but send a STOP in case it does
         throw new StopConditionException(e);
      }
      
      AppServerHelper helper = connect(config, p2jAccount,
                                       appServerName, sessionFree, sUser, sPwd, sInfo);
      
      // set the value returned by the CONNECT call.
      ControlFlowOps.setReturnValue(helper.getConnectResult());

      if (helper.getConnectError() != null)
      {
         // if there was an error, trhow it here.
         throw helper.getConnectError();
      }

      return helper;
   }

   /**
    * Get the appserver export local to the P2J server.  If not already created, obtain a local
    * instance by calling {@link RemoteObject#obtainLocalInstance}.
    * 
    * @return   The local appserver instance.
    */
   private static synchronized AppServerEntry getLocalAppServer()
   {
      if (localAppServer == null)
      {
         localAppServer = (AppServerEntry) RemoteObject.obtainLocalInstance(AppServerEntry.class,
                                                                            true);
      }
   
      return localAppServer;
   }

   /**
    * Send a STOP condition to the specified request.
    * 
    * @param    requestId
    *           The request where to send a STOP condition.
    */
   public void sendStop(int requestId)
   {
      appServer.sendStop(connectionID, requestId);
   }
   
   /**
    * Check if this connection is in session-free mode or in session-managed mode.
    * 
    * @return   The state of the {@link #sessionFree} flag.
    */
   public boolean isSessionFree()
   {
      return sessionFree;
   }
   
   /**
    * Get the ID of this appserver connection.
    *  
    * @return   See above.
    */
   public character getConnectionID()
   {
      return new character(connectionID);
   }
   
   /**
    * Disconnect from the appserver and terminate the associated P2J session.
    */
   public void disconnect()
   {
      appServer.disconnect(connectionID);
      if (session != null)
      {
         // terminate the session
         session.terminate();
      }
   }

   /**
    * Get the name of the remote procedure, identified by the given code.
    * 
    * @param    code
    *           The handle code of the remote procedure.
    *           
    * @return   The procedure's name.
    */
   public character getProcedureName(String code)
   {
      return appServer.getProcedureName(connectionID, code);
   }

   /**
    * Delete the associated remote procedure, from the remote side.
    * 
    * @param    code
    *           The handle code of the remote procedure.
    */
   public boolean deleteProcedure(String code)
   {
      return appServer.deleteProcedure(connectionID, code);
   }

   /**
    * Invoke a remote external procedure as persistent. The validation of the legacy program name
    * and arguments will be done on the remote side.
    * 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    h
    *           The handle where to save the proxy procedure. May be <code>null</code>.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   The result of this procedure/function invocation.
    */
   public character invokePersistent(character name,
                                     handle    h,
                                     boolean   transactionDistinct,
                                     Object... args)
   {
      return invokePersistent(name, h, transactionDistinct, null, args);
   }

   /**
    * Invoke a remote external procedure as persistent. The validation of the legacy program name
    * and arguments will be done on the remote side.
    * 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    h
    *           The handle where to save the proxy procedure. May be <code>null</code>.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   The result of this invocation.
    */
   public character invokePersistent(character name,
                                     handle    h,
                                     boolean   transactionDistinct,
                                     String    modes,
                                     Object... args)
   {
      return invokePersistent(null, name, h, transactionDistinct, modes, args);
   }

   /**
    * Invoke a remote remote user-defined function, using the given proxy procedure. The 
    * validation for the function name and arguments will be done on the remote side.
    * 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    h
    *           The proxy procedure in which to invoke the procedure or function.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   The result of this function invocation.
    */
   public BaseDataType invokeFunction(character name,
                                      handle    h,
                                      boolean   transactionDistinct,
                                      Object... args)
   {
      return invoke(name, h, true, false, false, transactionDistinct, null, args);
   }

   /**
    * Invoke a remote remote user-defined function, using the given proxy procedure. The 
    * validation for the function name and arguments will be done on the remote side.
    * 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    h
    *           The proxy procedure in which to invoke the procedure or function.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   The result of this function invocation.
    */
   public BaseDataType invokeFunction(character name,
                                      handle    h,
                                      boolean   transactionDistinct,
                                      String    modes,
                                      Object... args)
   {
      return invoke(name, h, true, false, false, transactionDistinct, modes, args);
   }

   /**
    * Invoke a remote remote user-defined function, using the given proxy procedure. The 
    * validation for the function name and arguments will be done on the remote side.
    * 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    h
    *           The proxy procedure in which to invoke the procedure or function.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   The result of this function invocation.
    */
   public BaseDataType invokeFunction(character name,
                                      handle    h,
                                      boolean   dynamicFunction,
                                      boolean   transactionDistinct,
                                      Object... args)
   {
      return invoke(name, h, true, dynamicFunction, false, transactionDistinct, null, args);
   }
   
   /**
    * Invoke a remote remote user-defined function, using the given proxy procedure. The 
    * validation for the function name and arguments will be done on the remote side.
    * 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    h
    *           The proxy procedure in which to invoke the procedure or function.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   The result of this function invocation.
    */
   public BaseDataType invokeFunction(character name,
                                      handle    h,
                                      boolean   dynamicFunction,
                                      boolean   transactionDistinct,
                                      String    modes,
                                      Object... args)
   {
      return invoke(name, h, true, dynamicFunction, false, transactionDistinct, modes, args);
   }

   /**
    * Invoke a remote external procedure or internal procedure. The validation of the legacy name
    * and arguments will be done on the remote side.
    * 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    h
    *           The proxy procedure in which to invoke the internal procedure.
    *           If <code>null</code>, an external procedure is invoked.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   The result of this procedure invocation.
    */
   public character invoke(character name,
                           handle    h,
                           boolean   transactionDistinct,
                           Object... args)
   {
      return new character(invoke(name, h, false, false, false, transactionDistinct, null, args));
   }

   /**
    * Invoke a remote external procedure or internal procedure. The validation of the legacy name
    * and arguments will be done on the remote side.
    * 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    h
    *           The proxy procedure in which to invoke the internal procedure.
    *           If <code>null</code>, an external procedure is invoked.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   The result of this procedure invocation.
    */
   public character invoke(character name,
                           handle    h,
                           boolean   transactionDistinct,
                           String    modes,
                           Object... args)
   {
      return new character(invoke(name, h, false, false, false, transactionDistinct, modes, args));
   }

   /**
    * Invoke a remote external procedure, internal procedure or user-defined function. The 
    * validation for the program name and arguments will be done on the remote side.
    * 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    h
    *           The proxy procedure in which to invoke the internal procedure or function.
    *           If <code>null</code>, an external procedure is invoked.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   The result of this procedure/function invocation.
    */
   public BaseDataType invoke(character name,
                              handle    h,
                              boolean   function,
                              boolean   dynamicFunction,
                              boolean   superCall,
                              boolean   transactionDistinct,
                              String    modes,
                              Object... args)
   {
      return invoke(null, name, h, 
                    function, dynamicFunction, superCall, transactionDistinct,
                    modes, args);
   }

   /**
    * Get the error returned by CONNECT call.
    * 
    * @return   See above.
    */
   public NumberedException getConnectError()
   {
      return this.connectError;
   }

   /**
    * Get the CONNECT call result.
    * 
    * @return   See above.
    */
   public character getConnectResult()
   {
      return this.connectResult;
   }
   
   /**
    * Get the name of the connected appserver.
    * 
    * @return   See above.
    */
   public String getAppServerName()
   {
      return this.appServerName;
   }

   /**
    * Initialize the specified proxy procedure.
    * 
    * @param    asyncReq
    *           The async request which refers an uninitialized proxy procedure 
    * @param    code
    *           The code of a persistent external procedure to which the agent is associated. May
    *           be <code>null</code>.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    */
   void initializeProxy(AsyncRequestImpl asyncReq,
                        String           code, 
                        boolean          transactionDistinct,
                        String           modes,
                        Object[]         args)
   {
      Object[] preprocessed = preProcessParameters(modes, args);

      // obtain an unique ID for this request
      int reqId = appServer.uniqueId();
      asyncReq.setRequestId(reqId);

      AppServerInvocationResult result = appServer.initializeProxy(connectionID, reqId, code, 
                                                                   transactionDistinct,
                                                                   modes, preprocessed);
      asyncReq.setResult(result);
   }

   /**
    * Obtain a proxy for the given external procedure. This will also reserve an agent which will
    * be used for all subsequent invocation (including external procedure initialization).
    * 
    * @param    name
    *           A name of a remote external procedure.
    *           
    * @return   A proxy to this procedure. May be invalid if the remote side can't resolve the
    *           procedure.
    */
   String obtainProxy(String name)
   {
      return appServer.obtainProxy(connectionID, name);
   }

   /**
    * Invoke a remote external procedure, internal procedure or user-defined function. The 
    * validation for the program name and arguments will be done on the remote side.
    * 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    h
    *           The proxy procedure in which to invoke the internal procedure or function.
    *           If <code>null</code>, an external procedure is invoked.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   The result of this procedure/function invocation.
    */
   BaseDataType invoke(AsyncRequestImpl asyncReq,
                       character        name,
                       handle           h,
                       boolean          function,
                       boolean          dynamicFunction,
                       boolean          superCall,
                       boolean          transactionDistinct,
                       String           modes,
                       Object...        args)
   {
      if (asyncReq != null && asyncReq._isCancelled())
      {
         // if the procedure is cancelled, nothing else to do
         asyncReq.setResult(new AppServerInvocationResult());
         return new unknown();
      }
   
      String pCode = null;
      if (h != null)
      {
         if (!ProcedureManager.isProxy(h))
         {
            throw new IllegalArgumentException("The passed handle is not a proxy procedure!");
         }
         
         pCode = (String) h.get();
      }
      
      // preprocess the parameters and save them in a distinct array, to be sent to the remote
      // side
      Object[] preprocessed = preProcessParameters(modes, args);
      
      // obtain an unique ID for this request
      int reqId = appServer.uniqueId();
      if (asyncReq != null)
      {
         asyncReq.setRequestId(reqId);
      }
      AppServerInvocationResult result = appServer.invoke(connectionID, reqId,
                                                          name, pCode,
                                                          function,dynamicFunction, superCall,
                                                          transactionDistinct,
                                                          modes, preprocessed);
      // if this is an async request, then no error messages are shown and errors are recorded 
      // before the EVENT-PROCEDURE is executed
      if (asyncReq == null)
      {
         checkResult(args, result);
      }
      else
      {
         asyncReq.setResult(result);
      }
   
      return (BaseDataType) result.getResult();
   }

   /**
    * Invoke a remote external procedure as persistent. The validation of the legacy program name
    * and arguments will be done on the remote side.
    * 
    * @param    asyncReq
    *           The {@link AsyncRequestImpl} instance when this is an async request, or null for
    *           sync requests.
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    h
    *           The handle where to save the proxy procedure. May be <code>null</code>.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   The result of this invocation.
    */
   character invokePersistent(AsyncRequestImpl asyncReq,
                              character        name,
                              handle           h,
                              boolean          transactionDistinct,
                              String           modes,
                              Object...        args)
   {
      if (asyncReq != null && asyncReq._isCancelled())
      {
         // if the procedure is cancelled, nothing else to do
         asyncReq.setResult(new AppServerInvocationResult());
         return new character("");
      }

      // preprocess the parameters and save them in a distinct array, to be sent to the remote
      // side
      Object[] preprocessed = preProcessParameters(modes, args);
      
      // obtain an unique ID for this request
      int reqId = appServer.uniqueId();
      if (asyncReq != null)
      {
         asyncReq.setRequestId(reqId);
      }
      AppServerInvocationResult result = appServer.invokePersistent(connectionID,
                                                                    reqId,
                                                                    name,
                                                                    transactionDistinct,
                                                                    modes,
                                                                    preprocessed);
      // if this is an async request, then no error messages are shown and errors are recorded 
      // before the EVENT-PROCEDURE is executed
      if (asyncReq == null)
      {
         checkResult(args, result);
      }
      else
      {
         asyncReq.setResult(result);
      }
      
      if (h != null)
      {
         character fileName = getProcedureName(result.getProcedureId());
         h.assign(new ProxyProcedureWrapper(result.getProcedureId(), fileName));
      }
      
      return new character((BaseDataType) result.getResult());
   }

   /**
    * Check the parameters before a remote call.  All OUTPUT or INPUT-OUTPUT parameters need to be
    * a P2J data wrapper, a {@link DataModelObject DMO}, {@link TableParameter} or a 
    * {@link TableResultSet result set}.
    * <p>
    * This will not change the references in the passed argument list. Instead, it will return a
    * distinct array which will need to be sent to the remote side.
    * 
    * @param    modes
    *           The parameter modes.  If <code>null</code>, no validation is performed.
    * @param    args
    *           The argument list.
    *           
    * @return   The preprocessed parameters.
    */
   private Object[] preProcessParameters(String modes, Object... args)
   {
      if (local)
      {
         // if the target appserver is in the same JVM as the requester, then do not process the
         // table parameters
         return args;
      }

      // the preprocessed parameter list
      Object[] params = new Object[args.length];

      final String msg =
         "Parameter %s in mode %s needs to be a P2J data type, table parameter or result set.";
      
      for (int i = 0; i < args.length; i++)
      {
         Object arg = args[i];

         char mode = (modes == null ? 0 : modes.charAt(i));

         if (arg instanceof TableParameter)
         {
            if (modes == null)
            {
               throw new IllegalArgumentException("Table parameters must have their mode set!");
            }

            TableParameter tableParam = (TableParameter) arg;
            Temporary tempDMO = tableParam.getTable();
            boolean append = tableParam.isAppend();
            
            boolean input  = (mode == 'I' || mode == 'U');
            boolean output = (mode == 'O' || mode == 'U');
            
            try
            {
               arg = new TableWrapper(new TempTableResultSet(tempDMO, input, output, append));
            }
            catch (ClassNotFoundException e)
            {
               String err = String.format("Could not process temp DMO %s as a parameter!", 
                                           tempDMO.getClass().getName());
               
               throw new IllegalArgumentException(err, e);
            }
         }
         else if (arg instanceof TableResultSet)
         {
            // make sure the remote side receives a known class
            arg = new TableWrapper((TableResultSet) arg);
         }
         else if (arg instanceof ResultSet)
         {
            // allow only input mode
            if (mode != 'I')
            {
               throw new IllegalArgumentException(
                  "java.sql.ResultSet instances are allowed only for INPUT table parameters!");
            }

            final String err = "Could not serialize the ResultSet on position " + i + ".";
            try
            {
               final ResultSet rs = (ResultSet) arg;
               final ResultSetMetaData rsmd = rs.getMetaData();
               final int columns = rs.getMetaData().getColumnCount();
               
               // build a custom TableResultSet which reads the data and metadata from the 
               // sql-like ResultSet instance, and wraps it in a TableWrapper for transport
               
               arg = new TableWrapper(new TableResultSet()
               {
                  private int nextCol = 0;

                  @Override
                  public boolean hasMoreRows()
                  {
                     try
                     {
                        return rs.next();
                     }
                     catch (SQLException e)
                     {
                        throw new IllegalArgumentException(err, e);
                     }
                  }

                  @Override
                  public Object[] nextRow()
                  {
                     try
                     {
                        Object[] row = new Object[columns];
                        
                        for (int i = 1; i <= columns; i++)
                        {
                           row[i] = rs.getObject(i);
                        }
                        
                        return row;
                     }
                     catch (SQLException e)
                     {
                        throw new IllegalArgumentException(err, e);
                     }
                  }
                  
                  @Override
                  public PropertyDefinition nextProperty()
                  {
                     try
                     {
                        // TODO: extent properties?
                        int type = rsmd.getColumnType(nextCol);
                        return new PropertyDefinition(rsmd.getColumnName(nextCol),
                                                      Persistence.toHibernateType(type));
                     }
                     catch (SQLException e)
                     {
                        throw new IllegalArgumentException(err, e);
                     }
                  }
                  
                  @Override
                  public boolean hasMoreProperties()
                  {
                     nextCol = nextCol + 1;
                     return nextCol <= columns;
                  }
               });
            }
            catch (SQLException e)
            {
               throw new IllegalArgumentException(err, e);
            }
         }
         else
         {
            // check if we have an array
            boolean isArray = arg.getClass().isArray();

            Object[] data = null;
            if (isArray)
            {
               // make sure we are going over each element in the array
               data = new Object[Array.getLength(arg)];
               for (int j = 0; j < data.length; j++)
               {
                  data[j] = Array.get(arg, j);
               }
            }
            else
            {
               data = new Object[] { arg };
            }
            
            for (int j = 0; j < data.length; j++)
            {
               Object darg = data[j];
               
               if (darg instanceof handle)
               {
                  handle h = (handle) darg;
                  if (h.isUnknown())
                  {
                     // if the handle is unknown, send a null resource ID
                     darg = new RemoteResourceImpl(null);
                  }
                  else
                  {
                     WrappedResource res = h.getResource();
                     if (res instanceof RemoteResource)
                     {
                        darg = res;
                     }
                     else
                     {
                        // any resource can be sent to the remote side, even if the remote side 
                        // can't resolve it properly
                        darg = new RemoteResourceImpl(h.toStringMessage());
                     }
                  }
               }
               else if (darg instanceof memptr)
               {
                  // do not allow memptrs to be used
                  throw new IllegalArgumentException(
                    "memptr instances can not be used from Java code.  Use MemoryBuffer " +
                    "instances instead.");
               }
               else if ((mode == 'U' || mode == 'O') && 
                        !(darg instanceof BaseDataType || darg instanceof MemoryBuffer))
               {
                  // for output/input-output modes, darg needs to be one of the known types
                  String err = String.format(msg, darg, (mode == 'U' ? "INPUT-OUTPUT" : "OUTPUT"));
                  throw new IllegalArgumentException(err);
               }
               
               data[j] = darg;
            }
            
            arg = (isArray ? data : data[0]);
         }
         
         params[i] = arg;
      }
      
      return params;
   }

   /**
    * Verify the current invocation result instance, by checking if any errors were thrown by the
    * remote call.
    * <p>
    * Also, this receives an argument list and updates it, depending on the parameter's mode 
    * (OUTPUT or INPUT-OUTPUT) and if the remote side sent an updated value of this argument. The
    * post-processing of the argument list is done only if no errors were received from the remote
    * side.
    * 
    * @param    args
    *           The arguments which need to be processed.
    * @param    result
    *           The result which needs to be checked for errors.
    */
   private void checkResult(Object[] args, AppServerInvocationResult result)
   {
      Throwable t = result.getError();
      if (t != null)
      {
         if (t instanceof NumberedException)
         {
            if (!SessionManager.get().isLeaf())
            {
               // do this only if we are in a P2J code
               NumberedException ne = (NumberedException) t;
               ErrorManager.displayError(ne.getNumber(), ne.getMessage(), false);
            }
            else
            {
               // we are not in P2J code, let the error propagate
               throw new RuntimeException(t);
            }

            Throwable cause = t.getCause();
            if (cause instanceof StopConditionException)
            {
               throw (ConditionException) cause;
            }
         }
         else if (t instanceof ConditionException)
         {
            throw (ConditionException) t;
         }
         else
         {
            throw new RuntimeException(t);
         }
      }
      
      // no errors found, post-process results
      postProcessResult(args, result);
   }

   /**
    * On the requester side, this receives an argument list and updates it, depending on the 
    * parameter's mode (OUTPUT or INPUT-OUTPUT) and if the remote side sent an updated value of
    * this argument.
    * 
    * @param    args
    *           The arguments (from the requester side) which need to be processed.
    * @param    result
    *           The result which needs to be checked for errors.
    */
   private void postProcessResult(Object[] args, AppServerInvocationResult result)
   {
      Object[] remoteArgs = result.getArguments();
      if (remoteArgs == null)
      {
         // some error happened and the remote side did not invoke anything or no arguments used
         return;
      }

      for (int i = 0; i < args.length; i++)
      {
         if (remoteArgs[i] == null)
         {
            // only output/input-output parameters are sent back
            continue;
         }
         
         Object localArg = args[i];
         Object remoteArg = remoteArgs[i];
         if (remoteArg instanceof TableWrapper)
         {
            // here there are two cases:
            // 1. from 4GL code, when a temp DMO is available
            // 2. from Java code, when a temp DMO is not available

            TableWrapper remoteResult = (TableWrapper) remoteArg;
            if (localArg instanceof TableParameter)
            {
               // DMO on the requester side, use the P2J persistence layer to update the  data
               TableParameter tableParam = (TableParameter) localArg;
               Temporary tempDMO = tableParam.getTable();
               boolean append = tableParam.isAppend();
               
               TemporaryBuffer.insertAllRows(remoteResult, tempDMO, append);
            }
            else if (localArg instanceof TableResultSet)
            {
               // we don't have a DMO, let the requester have access to the result set
               // the remoteArg reference must be pased to the local arg
               TableResultSet localResult = (TableResultSet) localArg;
               localResult.setRows(remoteResult.getRows());
            }
         }
         else
         {
            boolean isArray = localArg.getClass().isArray();
            Object[] remoteData = null;
            Object[] localData = null;
            
            if (isArray)
            {
               // check if this argument is an array
               remoteData = new Object[Array.getLength(remoteArg)];
               localData = new Object[Array.getLength(localArg)];
               
               for (int j = 0; j < remoteData.length; j++)
               {
                  remoteData[j] = Array.get(remoteArg, j);
               }
               for (int j = 0; j < localData.length; j++)
               {
                  localData[j] = Array.get(localArg, j);
               }
            }
            else
            {
               remoteData = new Object[] { remoteArg };
               localData = new Object[] { localArg };
            }
            
            if (remoteData.length != localData.length)
            {
               final String msg = 
                  "The remote side sent an array of length %d while the local argument is %d " +
                  "in length, for argument on position %d.";
               throw new IllegalArgumentException(String.format(msg,
                                                                remoteData.length,
                                                                localData.length,
                                                                i));
            }
            
            for (int j = 0; j < remoteData.length; j++)
            {
               Object dremoteArg = remoteData[j];
               Object dlocalArg = localData[j];

               if (dremoteArg instanceof BaseDataType)
               {
                  if (dlocalArg instanceof BaseDataType)
                  {
                     ((BaseDataType) dlocalArg).assign((BaseDataType) dremoteArg);
                  }
                  else
                  {
                     // requester's arg should have been a BDT, throw an error here but this 
                     // should never happen as preProcessParameters takes care to enforce this 
                     // constraint
                     final String msg = 
                       "Argument on index %d in OUTPUT or INPUT-OUTPUT mode must be of a P2J " +
                       "wrapper type.";
                     throw new IllegalArgumentException(String.format(msg, i));
                  }
               }
               else if (dlocalArg instanceof handle)
               {
                  handle h = (handle) dlocalArg;
                  RemoteResource res = (RemoteResource) dremoteArg;
                  if (res.unknown())
                  {
                     h.setUnknown();
                  }
                  else
                  {
                     h.assign(res);
                  }
               }
               else if (dlocalArg instanceof MemoryBuffer)
               {
                  MemoryBuffer remoteBuffer = (MemoryBuffer) dremoteArg;
                  MemoryBuffer localBuffer  = (MemoryBuffer) dlocalArg;
                  
                  localBuffer.setValue(remoteBuffer.getValue());
               }
            }
         }
      }
      
      // check if the result is a resource
      if (result.getResult() instanceof RemoteResource)
      {
         RemoteResource res = (RemoteResource) result.getResult();
         handle h = new handle();
         if (res.unknown())
         {
            h.setUnknown();
         }
         else
         {
            h.assign(res);
         }
         result.setResult(h);
      }
   }

   /**
    * Set the CONNECT call result.
    * 
    * @param    result
    *           The result return by the latest CONNECT call.
    */
   private void setConnectResult(character result)
   {
      this.connectResult = result;
   }

   /**
    * Set the error returned by the CONNECT call.
    * 
    * @param    error
    *           The error returned by the CONNECT call.
    */
   private void setConnectError(NumberedException error)
   {
      this.connectError = error;
   }
}
