/*
** Module   : AsyncRequestImpl.java
** Abstract : implementation of the async request resource.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130813 Created initial version.
*/

package com.goldencode.p2j.util;

import java.util.*;

import com.goldencode.p2j.NumberedException;
import com.goldencode.p2j.main.AppServerInvocationResult;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.util.*;

/**
 * Implements the async request resource. Beside adding support for its attributes, the async
 * resource is also responsible of initiating the async call and raising the PROCEDURE-COMPLETE
 * event.
 * <p>
 * As some the mutable fields of this class can be modified by more than one thread, access to 
 * them is synchronized.
 */
public class AsyncRequestImpl
extends HandleChain
implements AsyncRequest,
           Runnable
{
   /** A handle to the SERVER resource. */
   private final handle server = new handle();
   
   /** The name of the EVENT-PROCEDURE. */
   private final character eventProcedure = new character();
   
   /** The context of the EVENT-PROCEDURE. */
   private final handle eventProcedureContext = new handle();
   
   /** The proxy procedure, if the remote call is for an internal procedure. */
   private final handle persistentProcedure = new handle();
   
   /** The name of the target procedure. */
   private final character procedureName = new character();
   
   /** The procedure's arguments. */
   private final Object[] args;
   
   /** The modes for the procedure's arguments. */
   private String modes;
   
   /** The result from the remote call. */
   private AppServerInvocationResult result = null;
   
   /**
    * Flag indicating that the async call has completed (will be <code>false</code> if the call
    * was cancelled.
    */
   private boolean complete = false;
   
   /** Flag indicating that the async call has been cancelled. */
   private boolean cancelled = false;
   
   /** Flag indicating that the call was terminated with a QUIT condition. */
   private boolean quit = false;
   
   /** Flag indicating that the call was terminated with a STOP condition. */
   private boolean stop = false;

   /** Flag indicating that the call was terminated with an ERROR condition. */
   private boolean error = false;
   
   /** Flag indicating if the resource is valid. */
   private boolean valid = true;
   
   /** The ID of the remote request being executed. */
   private int requestId = -1;

   /**
    * List of listeners which need to receive notifications when the call is started or finished.
    */
   private List<AsyncRequestListener> listeners = new ArrayList<AsyncRequestListener>();

   /** A worker indicating what will be invoked. */
   private Runnable request;
   
   /**
    * Create a new async request resource, by specifying the attributes related to the async
    * call definition.
    * 
    * @param    server
    *           The server used to invoke the remote call. May be a <code>SERVER</code> or the
    *           <code>SESSION</code> resource.
    * @param    request
    *           The request which will execute the procedure call.
    * @param    eventProcedure
    *           The event procedure name.
    * @param    eventProcedureContext
    *           The context of the event procedure.
    * @param    procedureName
    *           The name of the procedure to be invoked.
    * @param    persistentProcedure
    *           The handle to the proxy procedure to which the target procedure belongs. Will
    *           be unknown if this is not a remote call.
    * @param    modes
    *           The modes for the procedure's arguments.
    * @param    args
    *           The procedure's arguments.
    */
   public AsyncRequestImpl(handle    server,
                           Runnable  request,
                           character eventProcedure,
                           handle    eventProcedureContext,
                           character procedureName,
                           handle    persistentProcedure,
                           String    modes,
                           Object... args)
   {
      this.server.assign(server);
      this.request = request;
      this.eventProcedure.assign(eventProcedure);
      this.eventProcedureContext.assign(eventProcedureContext);
      this.persistentProcedure.assign(persistentProcedure);
      this.procedureName.assign(procedureName);
      this.modes = modes;
      // arguments need special processing; copies are made of all BDT instances; TABLE
      // parameters remain intact
      this.args = new Object[args.length];
      for (int i = 0; i < args.length; i++)
      {
         if (args[i] instanceof BaseDataType)
         {
            this.args[i] = ((BaseDataType) args[i]).duplicate();
         }
      }
      
      setResourceType("ASYNC-REQUEST");
      setUnknown(false);
   }
   
   /**
    * Get the parameter modes for this async request.
    * 
    * @return   See above.
    */
   public String getModes()
   {
      return this.modes;
   }
   
   /**
    * Get the arguments for this async requests.
    * 
    * @return   See above.
    */
   public Object[] getArgs()
   {
      return this.args;
   }
   
   /**
    * Returns the state of the ERROR attribute
    *
    * @return    Current state of the ERROR attribute.
    */
   public synchronized logical isError()
   {
      return !complete ? new logical() : new logical(error);
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(boolean value)
   {
      readOnlyError("ERROR");
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(logical value)
   {
      readOnlyError("ERROR");
   }

   /**
    * Reports if this object is valid for use (has not been deleted).
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   @Override
   public boolean valid()
   {
      return valid;
   }

   /**
    * Returns a handle to the app server in which the procedure is ran.
    * 
    * @return   See above
    */
   @Override
   public handle getServerHandle()
   {
      return new handle(server);
   }

   /**
    * Get the <code>CANCELLED</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public synchronized logical isCancelled()
   {
      return new logical(cancelled);
   }

   /**
    * Get the <code>CANCELLED</code> attribute of this async request, as a Java boolean.
    * 
    * @return   See above.
    */
   public synchronized boolean _isCancelled()
   {
      return cancelled;
   }

   /**
    * Get the <code>COMPLETE</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public synchronized logical isComplete()
   {
      return new logical(complete);
   }

   /**
    * Get the <code>COMPLETE</code> attribute of this async request, as a Java boolean.
    * 
    * @return   See above.
    */
   public synchronized boolean _isComplete()
   {
      return complete;
   }

   /**
    * Get the <code>EVENT-PROCEDURE</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public character getEventProcedure()
   {
      return new character(eventProcedure);
   }

   /**
    * Get the <code>EVENT-PROCEDURE-CONTEXT</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public handle getEventProcedureContext()
   {
      return new handle(eventProcedureContext);
   }

   /**
    * Get the <code>PERSISTENT-PROCEDURE</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public handle getPersistentProcedure()
   {
      return new handle(persistentProcedure);
   }

   /**
    * Get the <code>PROCEDURE-NAME</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public character getProcedureName()
   {
      return new character(procedureName);
   }

   /**
    * Get the <code>QUIT</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public synchronized logical isQuit()
   {
      return new logical(quit);
   }

   /**
    * Get the <code>STOP</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public synchronized logical isStop()
   {
      return new logical(stop);
   }

   /**
    * Invoke the event procedure for this async request, if used.
    */
   public void run()
   {
      // change the RETURN-VALUE as set by this call
      Object res = result == null ? null : result.getResult();
      ControlFlowOps.setReturnValue(res == null ? null : new character((BaseDataType) res));

      // set the QUIT/STOP/ERROR flags and the details about any raised ERRORs
      Throwable cause = result.getError();
      while (cause != null)
      {
         if (cause instanceof QuitConditionException)
         {
            setQuit();
            break;
         }
         else if (cause instanceof StopConditionException)
         {
            setStop();
            break;
         }
         else if (cause instanceof ErrorConditionException)
         {
            setError();
            
            ErrorConditionException err = (ErrorConditionException) cause;
            if (err.getCause() instanceof NumberedException)
            {
               NumberedException ne = (NumberedException) err.getCause();
               ErrorManager.addError(ne.getNumber(), ne.getMessage(), true);
            }

            break;
         }

         cause = cause.getCause();
      }
      
      // no further processing is needed if no event procedure is set
      if (eventProcedure.isUnknown())
      {
         return;
      }

      List<Object> eventArgsList = new ArrayList<Object>();
      Object[] eventArgs = new Object[0];
      String eventModes = null;
      if (modes == null)
      {
         // get the resolved modes, if not explicitly passed at the call
         modes = result.getModes();
      }

      if (modes != null)
      {
         boolean failed = false;
         synchronized (this)
         {
            failed = error || quit || stop || cancelled;
         }

         for (int i = 0; i < modes.length(); i++)
         {
            char mode = modes.charAt(i);
            if (SourceNameMapper.isOutputMode(mode))
            {
               // if the request did not complete OK, pass an unknown, but only for BDTs
               eventArgsList.add(failed && args[i] instanceof BaseDataType ? new unknown() 
                                                                           : args[i]);
            }
         }
         
         eventModes = StringHelper.repeatChar(SourceNameMapper.INPUT_MODE, eventArgsList.size());
         eventArgs = eventArgsList.toArray();
      }
   
      // invoke the event procedure
      try
      {
         SelfManager.pushSelf(new handle(this));
         ControlFlowOps.invokeInWithMode(eventProcedure,
                                         eventProcedureContext,
                                         eventModes,
                                         eventArgs);
      }
      finally
      {
         SelfManager.popSelf();
      }
   }

   /**
    * Set the ID of this remote request.
    * 
    * @param    reqId
    *           The ID of the request.
    */
   synchronized void setRequestId(int reqId)
   {
      this.requestId = reqId;
   }

   /**
    * Called when the async call has finished normally (not with a cancel).
    */
   synchronized void complete()
   {
      this.complete = true;
      notifyListeners(false);
   }
   
   /**
    * Called when the async call was cancelled by the user.
    */
   synchronized void cancelled()
   {
      this.cancelled = true;
      notifyListeners(false);
   }

   /**
    * Mark this async call as having ended with a STOP condition.
    */
   synchronized void setStop()
   {
      this.stop = true;
   }

   /**
    * Mark this async call as having ended with a QUIT condition.
    */
   synchronized void setQuit()
   {
      this.quit = true;
   }

   /**
    * Mark this async call as having ended with a ERROR condition.
    */
   synchronized void setError()
   {
      this.error = true;
   }
   
   /**
    * Stop this async request by sending the STOP condition to the remote side. This will be
    * used only in cases when the async call is cancelled by the user.
    */
   synchronized void stop()
   {
      if (requestId == -1)
      {
         // the request was not yet started, prevent from starting it
         complete();
         setStop();
         return;
      }

      // send a STOP condition to the remote side
      ServerImpl s = (ServerImpl) server.get();
      s.getAppServerHelper().sendStop(requestId);

      // wait for it to complete
      while (!complete)
      {
         try
         {
            this.wait();
         }
         catch (InterruptedException e)
         {
            break;
         }
      }
   }

   /**
    * Set the result from the {@link #request} execution.
    * 
    * @param    result
    *           The result.
    */
   void setResult(AppServerInvocationResult result)
   {
      this.result = result;
   }
   
   /**
    * Register a new listener for notifications about the state of the async call.
    * 
    * @param    listener
    *           A listener instance.
    */
   void registerListener(AsyncRequestListener listener)
   {
      listeners.add(listener);
   }
   
   /**
    * Notify all listeners that the async call is about to start.
    */
   void notifyStart()
   {
      notifyListeners(true);
   }

   /**
    * Execute the async request. This will either raise the PROCEDURE-COMPLETE event or will
    * invoke the EVENT-PROCEDURE immediately.
    * 
    * @param    event
    *           Flag indicating if the PROCEDURE-COMPLETE event will be raised.
    */
   synchronized void execute(boolean event)
   {
      // check if the request was not cancelled (requests which are removed from queue and are
      // pending to be executed are marked as complete plus the STOP attribute to true
      if (complete)
      {
         return;
      }

      // execute the remote async call
      try
      {
         request.run();
      }
      catch (Throwable t)
      {
         // treat any conditions
         Throwable cause = t;
         while (cause != null && !(cause instanceof ConditionException))
         {
            cause = t.getCause();
         }
         
         if (cause != null)
         {
            if (cause instanceof StopConditionException)
            {
               setStop();
            }
            else if (cause instanceof QuitConditionException)
            {
               setQuit();
            }
            else if (cause instanceof ErrorConditionException)
            {
               setError();
            }
         }
      }
      finally
      {
         if (!_isCancelled())
         {
            complete();
         }

         if (event)
         {
            ServerEvent serverEvent = new ServerEvent(handle.resourceId(this),
                                                      Keyboard.PROCEDURE_COMPLETE,
                                                      this);
            LogicalTerminal.postServerEvent(serverEvent);
         }
         else
         {
            // execute the event now
            run();
         }
         
         // notify that the request has finished.
         synchronized (this)
         {
            this.notify();
         }
      }
   }
   
   /**
    * The async resource can be deleted only if it has been completed or cancelled.
    */
   @Override
   protected synchronized boolean resourceDelete()
   {
      if (!(complete || cancelled))
      {
         final String msg = "Asynchronous request  has not completed.  Cannot DELETE";
         ErrorManager.recordOrThrowError(8978, msg);
         return false;
      }
      
      this.valid = false;
      return true;
   }
   
   /**
    * As this resource supports the name attribute in R/W mode, return <code>false</code>
    * 
    * @return   always <code>false</code>
    */
   @Override
   protected boolean hasNameReadOnly()
   {
      return false;
   }
   
   /**
    * Notify the listeners about the invocation request state (start or finish).
    * 
    * @param    start
    *           <code>true</code> to indicate invocation start, <code>false</code> to notify for
    *           invocation finish.
    */
   private void notifyListeners(boolean start)
   {
      for (AsyncRequestListener listener : listeners)
      {
         if (start)
         {
            listener.notifyStart();
         }
         else
         {
            listener.notifyFinish();
         }
      }
   }
}
